/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * pathport.c
 * Part of libpathport
 * Copyright (C) 2006 Simon Newton
 */

// required for vasprintf
// patches welcomed to remove it :)
#define _GNU_SOURCE

#include "private.h"
#include <stdio.h>
#include <stdarg.h>

char *pathport_errstr = NULL;

/*
 * Libpathport error function
 * This writes the error string to pathport_strerror, which can be accessed
 * using pathport_strerror();
 *
 * Is vasprintf available on OSX ?
 */
void pathport_error(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);

  free(pathport_errstr);
  vasprintf(&pathport_errstr, fmt , ap);
  va_end(ap);
}


/*
 * Fetches the pathport error string
 *
 * @return the error string
 */
char *pathport_strerror() {
  return pathport_errstr;
}


/**
 * Creates a new Pathport node.
 * Takes a string containing the ip address of the interface to listen on
 * If the string is NULL it uses the first non loopback address
 *
 * @param ip the IP address to listen on
 * @param debug level of logging provided 0: none
 * @return an pathport_node
 */
pathport_node pathport_new(const char *ip, int debug) {
  node n;
  int ret;

  n = malloc(sizeof(pathport_node_t ) );

  if (n == NULL) {
    pathport_error("malloc failure");
    return NULL;
  }

  memset(n, 0x0, sizeof(pathport_node_t) );
  memset(n, 0x0, sizeof(n->dmx) );

  // default parameters
  n->state.mode = PATHPORT_OFF;
  n->state.verbose = debug;
  n->state.packet_count = 1;

  strncpy(n->state.name,"libpathport", PATHPORT_NAME_LENGTH);

  ret = pathport_net_init(n, ip);

  // Setup the pathport id, this needs to be globally unique
  // I can't think of a decent way to do this so we'll use the lower bytes of the
  // mac address
  n->state.pathport_id = PATHPORT_MANUF_ZP_TECH << 24 |  n->state.hw_addr[4] << 16 |
              n->state.hw_addr[5] << 8 | (n->state.hw_addr[3] % PathportMaxUni);

  if (ret) {
    free(n);
    return NULL;
  }

  return (pathport_node) n;
}

/**
 * Starts the Pathport node.
 * Binds to the network sockets
 *
 * @param vn the pathport_node
 * @return 0 on success, non 0 on failure
 */
int pathport_start(pathport_node vn) {
  node n = (node) vn;
  int ret;

  return_if_null(vn);
  return_if_on(n);

  if ((ret = pathport_net_start(n)) ) {
    return ret;
  }
  n->state.mode = PATHPORT_ON;

  return pathport_send_reply(n);
}


/*
 * Stops the Pathport node
 * Closes the network sockets held by the node
 *
 * @param vn the pathport_node
 * @return 0 on success, non 0 on failure
 */
int pathport_stop(pathport_node vn) {
  node n = (node) vn;
  int ret, i;

  return_if_null(vn);
  return_if_off(n);

  if ((ret = pathport_net_close(n) ))
    return ret;

  for (i = 0; i < PATHPORT_MAX_UNIVERSES; i++)
    if (n->dmx[i] != NULL)
      free(n->dmx[i]);


  n->state.mode = PATHPORT_OFF;

  return PATHPORT_EOK;
}


/*
 * Free the memory associated with this node
 *
 * @param vn the pathport_node
 */
int pathport_destroy(pathport_node vn) {
  return_if_null(vn);

  free(vn);
  return PATHPORT_EOK;
}


/*
 * Handle any received packets.
 * This function is the workhorse of libpathport. You have a couple of options:
 *   - use pathport_get_sd() to retrieve the socket descriptors and select to
 *     detect network activity. Then call pathport_read(node,0)
 *     when activity is detected.
 *   - call pathport_read repeatedly from within a loop with an appropriate
 *     timeout
 *
 * @param vn the pathport_node
 * @param timeout the number of seconds to block for if nothing is pending
 * @return 0 on success, -1 on failure
 */
int pathport_read(pathport_node vn, int timeout) {
  node n = (node) vn;
  pathport_ipacket_t p;
  int ret;

  return_if_null(vn);
  return_if_off(n);

  while(1) {
    memset( &p.p, 0x0, sizeof(p.p) );

    if ( (ret = pathport_net_recv(n, &p, timeout) ) < 0)
      return ret;

    // nothing to read
    if (ret == RECV_NO_DATA)
      break;

    // skip this packet (filtered)
    if (p.length == 0)
      continue;

    pathport_rx_handle(n, &p);
  }
  return PATHPORT_EOK;
}


/*
 * This is used to set handlers for sent/received pathport packets.
 * If you're using a stock standard node you more than likely don't want
 * to set these. See the pathport_set_dmx_callback and pathport_set_firmware_callback .
 * If you want to get down and dirty with pathport packets, you can set this
 * read / manipulate packets as they arrive (or get sent)
 *
 * @param vn     The pathport_node
 * @param handler  The handler to set
 * @param fh    A pointer to a function, set to NULL to turn off
 *             The function should return 0,
 * @param data    Data to be passed to the handler when its called
 *
 * @return 0 on sucess, non 0 on failure
 */
int pathport_set_handler(pathport_node vn, pathport_handler_name_t handler, int (*fh)(pathport_node n, void *pp, void * d), void *data) {
  node n = (node) vn;

  return_if_null(vn);

  if (handler < 0 || handler >= PATHPORT_LAST_HANDLER ) {
    pathport_error("%s : Invalid handler defined", __FUNCTION__ );
    return PATHPORT_EARG;
  }

  n->callbacks[handler].fh = fh;
  n->callbacks[handler].data = data;

  return PATHPORT_EOK;
}


/*
 * This is a special callback which is invoked when dmx data is received.
 * Unlike libpathport we pass the data back here, an esp node can recv on any
 * of 256 channels, we don't want to maintain buffer for them.
 *
 * @param vn     The pathport_node
 * @param fh    The callback to invoke (parameters passwd are the pathport_node, the port_id
 *           that received the dmx, and some user data
 * @param data    Data to be passed to the handler when its called
 */
int pathport_set_dmx_handler(pathport_node vn, int (*fh)(pathport_node n, unsigned int universe, unsigned int length, const uint8_t *data,  void *d), void *data ) {
  node n = (node) vn;

  return_if_null(vn);

  n->dmx_callback.fh = fh;
  n->dmx_callback.data = data;

  return PATHPORT_EOK;
}


/*
 * Send a pathport reply
 * This should be called every 6 seconds to stay compliant with the spec
 */
int pathport_send_reply(pathport_node vn) {
  node n = (node) vn;
  return_if_null(vn);
  return_if_off(n);

  int ret = pathport_tx_send_arprep(n, PATHPORT_ID_MCAST_ALL);

  if (ret) return ret;

  return pathport_tx_send_arprep(n, PATHPORT_ID_MCAST_MANAGE);
}


/*
 * Sends some dmx data
 *
 * @param vn the pathport_node
 * @param universe universe to address it to
 * @param length length of data
 * @param data pointer to dmx data
 *
 */
int pathport_send_dmx(pathport_node vn, unsigned int uni, uint16_t length, const uint8_t *data) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if ( length < 1 || length > PATHPORT_DMX_LENGTH) {
    pathport_error("%s : Length of dmx data out of bounds (should be between 1 and 512, was %i)", __FUNCTION__ , length);
    return PATHPORT_EARG;
  }

  if ( uni < 0 || uni > PathportMaxUni) {
    pathport_error("%s : Universe out of bounds (should be between 0 and %i, was %i)", __FUNCTION__ , PathportMaxUni, uni);
    return PATHPORT_EARG;
  }

  return pathport_tx_send_dmx(n, uni, length, data);
}


/*
 * Release a universe (sends a pathport release message)
 *
 * @param n the pathport node
 * @param uni the universe number to release
 */
int pathport_release_dmx(pathport_node vn, unsigned int uni) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if ( uni < 0 || uni >=  PathportMaxUni) {
    pathport_error("%s : Universe id out of bounds (should be between 0 and %i, was %i)",
      __FUNCTION__ , PathportMaxUni, uni);
    return PATHPORT_EARG;
  }

  return pathport_tx_send_dmx_release(n, uni);
}


/*
 * Sets the name of the node.
 * The string should be null terminated and a maxmium of 18 Characters will be used
 *
 * @param vn the pathport_node
 * @param name the short name of the node.
 */
int pathport_set_name(pathport_node vn, const char *name) {
  node n = (node) vn;

  return_if_null(vn);

  strncpy((char *) &n->state.name, name, PATHPORT_NAME_LENGTH);
  return PATHPORT_EOK;
}


/**
 * Sets the type of the node, this is only possible when the node is off
 *
 * @param vn the pathport_node
 * @param manu  the manufacturer id
 * @param class the pathport class
 * @param type the node type
 */
int pathport_set_type(pathport_node vn, pathport_manuf manu, pathport_class cls, pathport_class_type type) {
  node n = (node) vn;
  return_if_on(n);
  return_if_null(vn);

  n->state.pathport_id = manu << 24 | (n->state.pathport_id & 0x00ffffff );

  n->state.manuf = manu;
  n->state.cls = cls;
  n->state.type = type;

  return PATHPORT_EOK;
}


/*
 * Set the lower 3 bytes of the pathport id, this can only be done when the
 * node is off. If you call this you'd better understand how the ids are
 * assigned so as to avoid conflicts...
 *
 * @param vn the node
 * @param id the lower 3 bytes of the pathport id
 */
int pathport_set_id(pathport_node vn, uint32_t id) {
  node n = (node) vn;
  return_if_on(n);
  return_if_null(vn);

  n->state.pathport_id = (n->state.pathport_id & 0xff000000 ) | ( id & 0x00ffffff);
  return PATHPORT_EOK;
}


/*
 * Register our interest in this universe
 */
int pathport_register_uni(pathport_node vn, uint8_t uni) {
  node n = (node) vn;
  return_if_null(vn);

  if (uni >= PATHPORT_MAX_UNIVERSES)
    return PATHPORT_EARG;

  if (n->dmx[uni] == NULL) {
    n->dmx[uni] = (uint8_t*) malloc(PATHPORT_DMX_LENGTH * sizeof(uint8_t));

    if (n->dmx[uni] == NULL)
      return PATHPORT_EMEM;

    memset(n->dmx[uni], 0x00, PATHPORT_DMX_LENGTH);
  }
  printf("registered %i %p\n", uni, n->dmx[uni]);
  return 0;
}


/*
 * Unregister our interest in this universe
 */
int pathport_unregister_uni(pathport_node vn, uint8_t uni) {
  node n = (node) vn;
  return_if_null(vn);

  if (uni >= PATHPORT_MAX_UNIVERSES)
    return PATHPORT_EARG;

  if (n->dmx[uni] != NULL) {
    free(n->dmx[uni]);
    n->dmx[uni] = NULL;
  }

  return 0;
}


/*
 * Returns the socket descriptor associated with this pathport_node.
 * libpathport now uses a single descriptor per node  bound to the wildcard
 * address
 *
 * @param vn the pathport_node
 * @param socket the index of the socket descriptor to fetch (0 or 1 or 2)
 * @return the socket descriptor, or < 0 on error
 */
int pathport_get_sd(pathport_node vn, unsigned int i) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if ( i < PATHPORT_MAX_SD)
    return n->sd[i];
  else
    return n->sd[0];
}
