/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * transmit.c
 * Functions to handle sending datagrams
 * Copyright (C) 2006 Simon Newton
 */

#include "private.h"

/*
 * Build the header of a pathport packet
 *
 * @param n the node
 * @param p the internal representation of the packet
 * @param dst the destination pathport id
 */
void pathport_tx_build_packet(node n, pathport_ipacket p, pathport_id dst) {
  p->p.head.proto = htons(PATHPORT_PROTO);
  p->p.head.ver = htons(PATHPORT_VERSION);
  p->p.head.seq = htons(n->state.packet_count++);
  p->p.head.src = htonl(n->state.pathport_id);
  p->p.head.dst = htonl(dst);
}


/*
 * Send a pathport arp request
 *
 * @param n the pathport node
 * @param dst the destination id of the node (or PATHPORT_ID_BROADCAST)
 */
int pathport_tx_send_arpreq(node n, pathport_id dst) {
  pathport_ipacket_t p;
  pathport_packet_pdu *pdu = (pathport_packet_pdu*) p.p.data;

  memset(&p, 0x00, sizeof(p));

  p.to = n->state.status_addr;
  p.length = sizeof(pathport_packet_header) +
         sizeof(pathport_pdu_header) +
         sizeof(pathport_pdu_arpreq);

  pathport_tx_build_packet(n,&p, dst);

  pdu->head.type = htons(PATHPORT_ARPREQ);
  pdu->head.len = htons(sizeof(pathport_pdu_arpreq));

  return pathport_net_send(n, &p);
}


/*
 * Send a pathport arp reply
 * @param n the node
 * @param dst the destination id of the node (or PATHPORT_ID_BROADCAST)
 */
int pathport_tx_send_arprep(node n, pathport_id dst) {
  pathport_ipacket_t p;
  pathport_packet_pdu *pdu = (pathport_packet_pdu*) p.p.data;

  memset(&p, 0x00, sizeof(p));

  if ( dst == PATHPORT_ID_MCAST_MANAGE) {
    p.to = n->state.config_addr;
  } else {
    p.to = n->state.status_addr;
  }

  p.length = sizeof(pathport_packet_header) +
         sizeof(pathport_pdu_header) +
         sizeof(pathport_pdu_arprep);

  pathport_tx_build_packet(n,&p, dst);

  pdu->head.type = htons(PATHPORT_ARPREP);
  pdu->head.len = htons(sizeof(pathport_pdu_arprep));
  pdu->d.arprep.id = htonl(n->state.pathport_id);
  pdu->d.arprep.ip = n->state.ip_addr.s_addr;
  pdu->d.arprep.mcode = n->state.manuf;
  pdu->d.arprep.dclass = n->state.cls;
  pdu->d.arprep.dtype = n->state.type;
  pdu->d.arprep.num_co = 2;

  return pathport_net_send(n, &p);
}


/*
 * Send dmx data
 *
 * We segment the xDMX channel space into universes
 *
 * @param n the node
 * @param uni the universe (0 to PathportMaxUni)
 * @param length length of the dmx data
 * @param data pointer to the dmx data
 */
int pathport_tx_send_dmx(node n, uint8_t uni, int length, const uint8_t *data) {
  pathport_ipacket_t p;
  pathport_packet_pdu *pdu = (pathport_packet_pdu*) p.p.data;

  memset(&p, 0x00, sizeof(p));

  p.to = n->state.data_addr;
  p.length = sizeof(pathport_packet_header) +
         sizeof(pathport_pdu_header) +
         sizeof(pathport_pdu_data) + length;

  pathport_tx_build_packet(n,&p, PATHPORT_ID_BROADCAST);

  pdu->head.type = htons(PATHPORT_DATA);
  pdu->head.len = htons(sizeof(pathport_pdu_data) + length);
  pdu->d.data.type = htons(PATHPORT_XDMX_DATA_FLAT);
  pdu->d.data.num = htons(length);
  pdu->d.data.offset = htons(uni * PATHPORT_DMX_LENGTH);
  memcpy(pdu->d.data.data , data, length);
  return pathport_net_send(n, &p);
}


/*
 * Sends a get reply to dst with the specified params
 *
 * @param n the node
 * @param dst the pathport id of the destination
 * @param params list of the params to send
 * @param num number of params
 */
int pathport_tx_send_getreply(node n, pathport_id dst, pathport_param *params, unsigned int num) {
  pathport_ipacket_t p;
  pathport_packet_pdu *pdu = (pathport_packet_pdu*) p.p.data;
  int i, l;
  int offset = 0;
  uint16_t param;
  pathport_pdu_getrep_alv *a;

  memset(&p, 0x00, sizeof(p));

//  if ( dst == PATHPORT_ID_MCAST_MANAGE) {
    p.to = n->state.config_addr;
//  } else {
//    p.to = n->state.status_addr;
//  }

  pathport_tx_build_packet(n,&p, dst);
  pdu->head.type = htons(PATHPORT_GETREP);

  a = (pathport_pdu_getrep_alv*) pdu->d.getrep.params;

// check for buffer overrun here
  for (i=0; i<num; i++) {
    param = ntohs(params[i]);
    switch (param) {
      case PATHPORT_PARAM_NODE_NAME:
        printf("match %x\n", param);
        a->len = strlen(n->state.name);
        memcpy(a->val, n->state.name, a->len );
        break;
      case PATHPORT_PARAM_MAC_ADDR:
        printf("match %x\n", param);
        a->len = 6;  // ok
        memcpy(a->val, n->state.hw_addr, 6);
        break;
      case PATHPORT_PARAM_IP_ADDR:
        printf("match %x\n", param);
        a->len = 4;   // ok
        memcpy( a->val, &n->state.ip_addr, 4);
        break;
      case PATHPORT_PARAM_SUBNET_MASK:
        printf("match %x\n", param);
        a->len = 4;   // ok
        memcpy( a->val, &n->state.ip_addr, 4);
        break;
      case PATHPORT_PARAM_GATEWAY:
        printf("match %x\n", param);
        a->len = 4;  // ok
        memcpy( a->val, &n->state.ip_addr, 4);
        break;
      case PATHPORT_PARAM_PATHPORT_ID :
        printf("match %x\n", param);
        a->len = 4;
        memcpy( a->val, &n->state.pathport_id, 4);
        break;
      case PATHPORT_PARAM_BACKLIGHT:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SW_VERSION:
        printf("match %x\n", param);
        a->len = 4;
        *a->val = 0x0a;
        break;
      case PATHPORT_PARAM_HW_TYPE:
        printf("match %x\n", param);
        a->len = 4;
        *a->val = 0x0a;
        break;
      case PATHPORT_PARAM_LDR_VERSION:
        printf("match %x\n", param);
        a->len = 4;
        *a->val = 0x0b;
        break;
      case PATHPORT_PARAM_IDENTIFY:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_IR_ENABLE:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SERIAL:
        printf("match %x\n", param);
        a->len = 4;
        *a->val = 0x12;
        break;
      case PATHPORT_PARAM_BOOT_ORDER:
        printf("match %x\n", param);
        a->len = 4;      // ok
        *a->val = 0;
        a->val[2] = 0x05;
        a->val[3] = 0x01;
        break;
      case PATHPORT_PARAM_TERMINAL_CONNECTORS:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 0;
        break;
      case PATHPORT_PARAM_RACK_MOUNT:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 0;
        break;
      case PATHPORT_PARAM_BOX_STYLE:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 1;
        break;







      case PATHPORT_PARAM_COMPONENT_LABEL:
        printf("match %x\n", param);
        a->len = 4;  // ok
        memcpy(a->val , "test", 4);
        break;
      case PATHPORT_PARAM_PATCH_NAME:
        printf("match %x\n", param);
        a->len = 9;
        memcpy(a->val , "Availably" , 9);
        break;
      case PATHPORT_PARAM_XDMX_SRC_OFFSET:
        printf("match %x\n", param);
        a->len = 2;  // ok
        *a->val = 0x02;
        break;
      case PATHPORT_PARAM_ALT_PATCH_MODE_ENABLE:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0x00;
        break;
      case PATHPORT_PARAM_QUICK_PATCH_UNI:
        printf("match %x\n", param);
        a->len = 2;
        *a->val = 0;
        break;
      case PATHPORT_PARAM_DMX_SPEED:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_IS_BIDIRECTIONAL:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 1;
        break;
      case PATHPORT_PARAM_IS_PHYSICAL:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 1;
        break;
      case PATHPORT_PARAM_IS_MALE:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_IS_SINK:
        printf("match %x\n", param);
        a->len = 1;   // ok
        *a->val = 1;
        break;
      case PATHPORT_PARAM_XDMX_COUNT:
        printf("match %x\n", param);
        a->len = 2;   // ok
        *a->val = 0x02;
        break;
      case PATHPORT_PARAM_ALT_START_CODE:
        printf("match %x\n", param);
        a->len = 4; // ok
        break;
      case PATHPORT_PARAM_MAX_PATCHES:
        printf("match %x\n", param);
        a->len = 2; // ok
        *a->val = 0x10;
        break;
      case PATHPORT_PARAM_NUM_PATCHES:
        printf("match %x\n", param);
        a->len = 2;   // ok
        *a->val = 0x00;
        break;
      case PATHPORT_PARAM_TALKBACK:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 0x00;
        break;



      case PATHPORT_PARAM_PORT_ENABLE:
        printf("match %x\n", param);
        a->len = 1;     // ok
        *a->val = 1;
        break;
      case PATHPORT_PARAM_DMX_ACTIVE:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 1;
        break;
      case PATHPORT_PARAM_XDMX_ACTIVE:
        printf("match %x\n", param);
        a->len = 1;
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SIG_LOSS_HOLD:
        printf("match %x\n", param);
        a->len = 4;    // ok
        *a->val = 0x00;
        break;

      case PATHPORT_PARAM_SIG_LOSS_HOLD_FOREVER:
        printf("match %x\n", param);
        a->len = 1;     // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SIG_LOSS_FADE_ENABLE:
        printf("match %x\n", param);
        a->len = 1; // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SIG_LOSS_FADE_TIME:
        printf("match %x\n", param);
        a->len = 4; // ok
        *a->val = 0;
        break;
      case PATHPORT_PARAM_SIG_LOSS_PORT_SHUTDOWN:
        printf("match %x\n", param);
        a->len = 1;   //ok
        *a->val = 0x00;
        break;

      default:
        // do nothing
        printf("no match %x\n" , param );

    }

    if (a->len) {
      // pad to a multiple of 4 bytes if needed
      if (a->len % 4 != 0)
        l =  a->len + 4 - (a->len %4);
      else
        l = a->len;

      offset += (sizeof(pathport_pdu_getrep_alv) + l);
      a->type = params[i];
      a->len = htons(a->len);
      a = (pathport_pdu_getrep_alv*) &pdu->d.getrep.params[offset];
    }
  }


  p.length = sizeof(pathport_packet_header) +
         sizeof(pathport_pdu_header) +
         sizeof(pathport_pdu_getrep) + offset;
printf("length is %i %i\n" , offset , p.length);

  pdu->head.len = htons(sizeof(pathport_pdu_getrep) + offset);

  return pathport_net_send(n, &p);
}


/*
 * Send a pathport dmx release message
 *
 * @param n the node
 * @param uni the universe ( 0 to PathportMaxUni)
 */
int pathport_tx_send_dmx_release(node n, int uni) {
  pathport_ipacket_t p;
  pathport_packet_pdu *pdu = (pathport_packet_pdu*) p.p.data;

  memset(&p, 0x00, sizeof(p));

  p.to = n->state.data_addr;
  p.length = sizeof(pathport_packet_header) +
         sizeof(pathport_pdu_header) +
         sizeof(pathport_pdu_data);

  pathport_tx_build_packet(n,&p, PATHPORT_ID_MCAST_DATA);

  pdu->head.type = htons(PATHPORT_DATA);
  pdu->head.len = htons(sizeof(pathport_pdu_data));
  pdu->d.data.type = PATHPORT_XDMX_DATA_RELEASE;
  pdu->d.data.num = PATHPORT_DMX_LENGTH;
  pdu->d.data.offset = uni * PATHPORT_DMX_LENGTH;
  return pathport_net_send(n, &p);
}
